Package: controller
Title: Controlled Vocabularies
Version: 0.1.0
Authors@R: 
    person(given = "Joe",
           family = "Roe",
           role = c("aut", "cre", "cph"),
           email = "joe@joeroe.io",
           comment = c(ORCID = "0000-0002-1011-1244"))
Description: Work with controlled vocabularies; a predefined set of terms 
  intended to be used in in a specific context. This package introduces the 
  'control' verb, which recodes values in a vector using a lookup table of 
  preferred and variant terms (a 'thesaurus'). It also includes functions for 
  reading controlled vocabularies from standard formats.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Imports: curl, dplyr, fs, rlang, utils
Depends: R (>= 4.1.0)
URL: https://controller.joeroe.io, https://github.com/joeroe/controller
BugReports: https://github.com/joeroe/controller/issues
Suggests: covr, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-07-16 03:55:18 UTC; joeroe
Author: Joe Roe [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-1011-1244>)
Maintainer: Joe Roe <joe@joeroe.io>
Repository: CRAN
Date/Publication: 2026-07-23 13:30:07 UTC

# Generated by roxygen2: do not edit by hand

export(control)
export(control_ci)
export(control_fuzzy)
export(control_matches)
export(control_names)
export(control_names_ci)
export(control_names_fuzzy)
export(read_fish)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# controller

<!-- badges: start -->

[![Project Status: WIP – Initial development is in progress, but there
has not yet been a stable, usable release suitable for the
public.](https://www.repostatus.org/badges/latest/wip.svg)](https://www.repostatus.org/#wip)
[![R-CMD-check](https://github.com/joeroe/controller/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/joeroe/controller/actions/workflows/R-CMD-check.yaml)
[![Test
coverage](https://codecov.io/gh/joeroe/controller/graph/badge.svg)](https://app.codecov.io/gh/joeroe/controller)
<!-- badges: end -->

**controller** is a collection of functions for working with controlled
vocabularies in R. It introduces the `control()` verb, which recodes
values in a vector using a lookup table of preferred and variant terms
(a *thesaurus*).

## Installation

You can install the latest release of controller [from
CRAN](https://cran.r-project.org/package=controller) with:

``` r
install.packages("controller")
```

Or the development version from GitHub using the
[remotes](https://remotes.r-lib.org/) package:

``` r
# install.packages("remotes")
remotes::install_github("joeroe/controller")
```

## Usage

A common data-tidying problem is standardising variant terms for the
same concept. Imagine we have a dataset that uses a number of different
names for shades of the same colour. As data analysts, we naturally want
to recode the data to eliminate this messy creativity, for example using
[dplyr::recode()](https://dplyr.tidyverse.org/reference/recode.html):

``` r
library(dplyr, warn.conflicts = FALSE)
shades <- c("daffodil", "purple", "magenta", "azure", "navy", "violet")

recode(shades,
       daffodil = "yellow",
       purple = "purple",
       magenta = "pink",
       azure = "blue",
       navy = "blue",
       violet = "purple")
#> [1] "yellow" "purple" "pink"   "blue"   "blue"   "purple"
```

But recoding this way can be tedious, especially if there are a large
number of terms. With `control()`, we can instead use a data frame
containing a thesaurus to replace the values:

``` r
library(controller)
data("colour_thesaurus")

control(shades, colour_thesaurus)
#> Replaced values:
#> ℹ daffodil → yellow
#> ℹ azure → blue
#> ℹ navy → blue
#> ℹ violet → purple
#> Warning: Some values of `x` were not matched in `thesaurus`:
#> ✖ magenta
#> [1] "yellow"  "purple"  "magenta" "blue"    "blue"    "purple"
```

`control()` also supports fuzzy matching, removing the need to
exhaustively list variants for common causes of differing terminology.
For example, to perform a case insensitive match to the thesaurus:

``` r
shades <- toupper(shades)
control_ci(shades, colour_thesaurus)
#> Replaced values:
#> ℹ DAFFODIL → yellow
#> ℹ PURPLE → purple
#> ℹ AZURE → blue
#> ℹ NAVY → blue
#> ℹ VIOLET → purple
#> Warning: Some values of `x` were not matched in `thesaurus`:
#> ✖ MAGENTA
#> [1] "yellow"  "purple"  "MAGENTA" "blue"    "blue"    "purple"
```

# control.R
# Core verb `control()`

#' Recode values using a thesaurus
#'
#' @description
#' The `control()` verb replaces values in a vector with values looked up in a
#' thesaurus. It is similar to [switch()] or [dplyr::recode()], but the
#' replacement values are specified as a data frame instead of as individual
#' arguments.
#'
#' By default `control()` replaces only values of `x` that exactly match terms
#' in `thesaurus`. Additional arguments allow for case insensitive and fuzzy
#' matching strategies (see details). `control_ci()` and `control_fuzzy()` are
#' convenience aliases for case insensitive exact matching and full fuzzy
#' matching respectively.
#'
#' @param x Vector to recode.
#' @param thesaurus Data frame with a vector of preferred terms and a vector of
#'   variants.
#' @param thesaurus_cols Vector of two column names or positions specifying which
#'   columns in `thesaurus` contain the preferred terms and variants
#'   respectively. Defaults to the first two columns.
#' @param case_insensitive Set to `TRUE` to perform case insensitive matching.
#' @param fuzzy_boundary Set to `TRUE` to perform fuzzy matching that ignores
#'   differences in the word boundaries used (e.g. `"foo bar"` matches `"foo-bar"`).
#' @param fuzzy_encoding Set to `TRUE` to perform fuzzy matching that ignores
#'   non-ASCII characters that may have been encoded differently (e.g. `"foo"`
#'   matches `"foö"`).
#' @param quiet Set to `TRUE` suppress messages about replaced values.
#' @param warn_unmatched If `TRUE` (the default), issues a warning for values
#'   that couldn't be matched in `thesaurus`.
#' @param ... For `control_ci()` and `control_fuzzy()`, other arguments passed
#'   to `control()`. This includes fuzzy matching options (`fuzzy_boundary`,
#'   `fuzzy_encoding`) and output options (`quiet`, `warn_unmatched`).
#'
#' @return
#' A vector the same length as `x` with values matching variants in `thesaurus`
#' replaced with the preferred term. NAs in `x` are preserved as NAs.
#'
#' By default gives a message listing replaced values and a warning listing any
#' values not matched in the thesaurus. These can be suppressed with
#' `quiet = TRUE` and `warn_unmatched = FALSE` respectively.
#'
#' @export
#'
#' @examples
#' data(colour_thesaurus)
#'
#' # Exact matching
#' x <- c("red", "lipstick", "green", "mint", "blue", "azure")
#' control(x, colour_thesaurus)
#'
#' # Case insensitive matching
#' x <- toupper(x)
#' control_ci(x, colour_thesaurus)
#'
#' # control_matches() returns a data frame showing which match type was used:
#' control_matches(x, colour_thesaurus, case_insensitive = TRUE)
control <- function(x, thesaurus,
                    thesaurus_cols = c(1, 2),
                    case_insensitive = FALSE,
                    fuzzy_boundary = FALSE,
                    fuzzy_encoding = FALSE,
                    quiet = FALSE,
                    warn_unmatched = TRUE) {
  if (!is.vector(x)) {
    rlang::abort("`x` must be a vector.")
  }

  terms <- unique(x[!is.na(x)])
  thesaurus <- prepare_thesaurus(thesaurus, thesaurus_cols)

  matched <- match_terms(terms, thesaurus, case_insensitive,
                         fuzzy_boundary, fuzzy_encoding)
  matched$match <- do.call(dplyr::coalesce, c(matched[, -1, drop = FALSE],
                                              matched["term"]))

  # TODO: detect ambiguous matches and warn or error (#4)

  # Inform
  if (isFALSE(quiet)) {
    replaced <- matched[matched$term != matched$match, , drop = FALSE]
    if (nrow(replaced) > 0) {
      replaced <- paste(replaced$term, "\u2192", replaced$match)
      names(replaced) <- rep("i", length(replaced))
      rlang::inform(c("Replaced values:", replaced))
    }
    else {
      rlang::inform("No values replaced.")
    }
  }

  if (isTRUE(warn_unmatched)) {
    match_types <- matched[, !colnames(matched) %in% c("term", "match"),
                           drop = FALSE]
    unmatched <- matched$term[rowSums(is.na(match_types)) == ncol(match_types)]

    if (length(unmatched) > 0) {
      names(unmatched) <- rep("x", length(unmatched))
      rlang::warn(
        c("Some values of `x` were not matched in `thesaurus`:", unmatched),
        class = "controller_incomplete_control"
      )
    }
  }

  result <- matched$match[match(x, matched$term)]
  result[is.na(x)] <- NA
  result
}

#' @rdname control
#' @export
control_ci <- function(x, thesaurus, thesaurus_cols = c(1, 2), ...) {
  control(x, thesaurus, thesaurus_cols, case_insensitive = TRUE, ...)
}

#' @rdname control
#' @export
control_fuzzy <- function(x, thesaurus, thesaurus_cols = c(1, 2), ...) {
  control(x, thesaurus, thesaurus_cols, case_insensitive = TRUE,
          fuzzy_boundary = TRUE, fuzzy_encoding = TRUE, ...)
}

test_that("control() aborts if `x` is not a vector", {
  expect_error(control(data.frame(), colour_thesaurus),
               regexp = "`x` must be a vector")
})

test_that("control() with exact matching works", {
  expect_equal(control(colour_thesaurus$shade, colour_thesaurus, quiet = TRUE),
               colour_thesaurus$colour)
})

test_that("control() with case insensitive matching works", {
  expect_equal(control_ci(toupper(colour_thesaurus$shade), colour_thesaurus, quiet = TRUE),
               colour_thesaurus$colour)
})

test_that("control() issues correct messages and warnings", {
  x <- c("a", "b", "c", "x", "y", "z")
  y <- data.frame(c("X", "Y", "Z"), c("x", "y", "z"))
  z <- c(data.frame(c("a", "b", "c", "x", "y", "z"),
                    c("a", "b", "c", "x", "y", "z")))

  # Replacement message, unmatched warning
  expect_message(control(x, y, warn_unmatched = FALSE), regexp = "Replaced values")
  expect_warning(control(x, y, quiet = TRUE), class = "controller_incomplete_control")

  # No replacement message, no warning
  expect_message(control(x, z), regexp = "No values replaced")
  expect_warning(control(x, z, quiet = TRUE), regexp = NA)

  # Suppressed message + warning
  expect_message(control(x, y, warn_unmatched = FALSE, quiet = TRUE), regexp = NA)
  expect_warning(control(x, y, warn_unmatched = FALSE, quiet = TRUE), regexp = NA)
})

test_that("control() preserves NAs in x", {
  x <- c("red", NA, "blue")
  result <- control(x, colour_thesaurus, quiet = TRUE)

  expect_equal(result, c("red", NA, "blue"))
  expect_equal(is.na(result), c(FALSE, TRUE, FALSE))
})

test_that("control() returns unmatched terms as-is", {
  x <- c("red", "unknown", "blue")
  result <- control(x, colour_thesaurus, quiet = TRUE, warn_unmatched = FALSE)
  expect_equal(result, c("red", "unknown", "blue"))
})

test_that("control() and control_matches() produce consistent results", {
  x <- c("red", "lipstick", "green", "unknown")

  # Exact matching only
  ctrl_result <- control(x, colour_thesaurus, quiet = TRUE, warn_unmatched = FALSE)
  matches_result <- control_matches(x, colour_thesaurus)
  coalesced <- dplyr::coalesce(matches_result$exact_match, matches_result$term)
  expect_equal(coalesced, ctrl_result)

  # Case insensitive matching
  x_ci <- toupper(x)
  ctrl_result_ci <- control_ci(x_ci, colour_thesaurus, quiet = TRUE, warn_unmatched = FALSE)
  matches_result_ci <- control_matches(x_ci, colour_thesaurus, case_insensitive = TRUE)
  coalesced_ci <- dplyr::coalesce(matches_result_ci$exact_match,
                                  matches_result_ci$case_insensitive_match,
                                  matches_result_ci$term)
  expect_equal(coalesced_ci, ctrl_result_ci)
})

test_that("control() does not warn about unmatched NAs", {
  expect_warning(control(c("red", NA), colour_thesaurus, quiet = TRUE), regexp = NA)
})

test_that("control() accepts thesaurus_cols by name", {
  df <- data.frame(colour = c("red", "green"), shade = c("crimson", "mint"))
  expect_equal(control(c("crimson", "mint"), df, thesaurus_cols = c("colour", "shade"), quiet = TRUE),
               c("red", "green"))
})

test_that("control() accepts thesaurus_cols by position", {
  df <- data.frame(colour = c("red", "green"), shade = c("crimson", "mint"))
  expect_equal(control(c("crimson", "mint"), df, thesaurus_cols = c("colour", "shade"), quiet = TRUE),
               control(c("crimson", "mint"), df, thesaurus_cols = c(1, 2), quiet = TRUE))
})

test_that("control() errors if thesaurus doesn't have exactly 2 columns", {
  df <- data.frame(a = 1:3, b = 1:3, c = 1:3)
  expect_error(control(1:3, df, thesaurus_cols = 1:3, quiet = TRUE), "`thesaurus_cols` must specify exactly 2 columns")
})

test_that("control() errors if variants are not unique", {
  df <- data.frame(a = c("x", "x"), b = c("y", "y"))
  expect_error(control("y", df, quiet = TRUE), "Variants .* must be unique")
})

test_that("control() with fuzzy boundary matching works", {
  df <- data.frame(canon = "foo bar", variant = "foo-bar")
  expect_equal(
    control(c("foo bar", "foo_bar", "foobar"), df,
            fuzzy_boundary = TRUE, quiet = TRUE, warn_unmatched = FALSE),
    c("foo bar", "foo bar", "foo bar")
  )
})

test_that("control() with fuzzy encoding matching works", {
  df <- data.frame(canon = "bar", variant = "fo\u00f6")
  expect_equal(
    control("foo", df, fuzzy_encoding = TRUE, quiet = TRUE, warn_unmatched = FALSE),
    "bar"
  )
})

test_that("control_fuzzy() combines all matching strategies", {
  boundary <- data.frame(canon = "foo bar", variant = "foo-bar")
  encoding <- data.frame(canon = "baz", variant = "ba\u00e7")

  expect_equal(
    control_fuzzy(c("foo bar", "foo_bar", "bac"), rbind(boundary, encoding),
                  quiet = TRUE, warn_unmatched = FALSE),
    c("foo bar", "foo bar", "baz")
  )
})

test_that("control_ci() accepts fuzzy matching arguments", {
  df <- data.frame(canon = "foo bar", variant = "foo-bar")
  expect_equal(
    control_ci(c("foo bar", "foo_bar"), df,
               fuzzy_boundary = TRUE, quiet = TRUE, warn_unmatched = FALSE),
    c("foo bar", "foo bar")
  )
})

test_that("control_ci() defaults to exact matching only", {
  df <- data.frame(canon = "foo bar", variant = "foo-bar")
  expect_equal(
    control_ci(c("foo bar", "foo_bar"), df,
               quiet = TRUE, warn_unmatched = FALSE),
    c("foo bar", "foo_bar")
  )
})

# utils.R
# Miscellaneous internal utility functions

# Is a path a URL?
# Copied from readr:::is_url()
is_url <- function(x) {
  grepl("^((http|ftp)s?|sftp)://", x)
}

# Is a file a zip archive?
# Checks magic bytes (PK signature: 50 4B 03 04)
is_zip <- function(path) {
  if (!fs::file_exists(path) || fs::is_dir(path)) return(FALSE)
  bytes <- readBin(path, "raw", n = 4)
  length(bytes) >= 4 &&
    bytes[1] == as.raw(0x50) &&
    bytes[2] == as.raw(0x4B) &&
    bytes[3] == as.raw(0x03) &&
    bytes[4] == as.raw(0x04)
}

test_that("is_url() identifies valid URLs", {
  expect_true(is_url("http://example.com"))
  expect_true(is_url("https://example.com"))
  expect_true(is_url("ftp://example.com/file"))
  expect_true(is_url("sftp://example.com/file"))
})

test_that("is_url() rejects non-URLs", {
  expect_false(is_url("example.com"))
  expect_false(is_url("/path/to/file"))
  expect_false(is_url("not a url"))
  expect_false(is_url(""))
})

test_that("is_url() is vectorized", {
  urls <- c("https://example.com", "not a url", "ftp://test.org", "/local/path")
  result <- is_url(urls)
  expect_equal(result, c(TRUE, FALSE, TRUE, FALSE))
})

test_that("is_zip() identifies zip files", {
  tmp <- tempfile(fileext = ".zip")
  on.exit(unlink(tmp))

  # Create a minimal zip file using base R's zip()
  src <- tempfile(fileext = ".txt")
  writeLines("test content", src)
  zip(tmp, src)
  unlink(src)

  expect_true(is_zip(tmp))
})

test_that("is_zip() rejects non-zip files", {
  tmp <- tempfile(fileext = ".txt")
  on.exit(unlink(tmp))

  writeLines("not a zip file", tmp)
  expect_false(is_zip(tmp))
})

test_that("is_zip() rejects non-existent files", {
  expect_false(is_zip("/nonexistent/file/path.zip"))
})

test_that("is_zip() rejects directories", {
  tmp <- tempfile()
  on.exit(unlink(tmp, recursive = TRUE))

  dir.create(tmp)
  expect_false(is_zip(tmp))
})

test_that("is_zip() rejects files too small to be zips", {
  tmp <- tempfile()
  on.exit(unlink(tmp))

  writeBin(as.raw(c(0x50, 0x4B)), tmp)  # Only 2 bytes, need 4
  expect_false(is_zip(tmp))
})

# control_matches.R
# Inspect detailed match results

#' Show detailed match results from a thesaurus
#'
#' @description
#' `control_matches()` returns a data frame showing which type of match was used
#' for each value in `x`. This is useful for debugging or inspecting how
#' `control()` recodes values.
#'
#' @inheritParams control
#'
#' @return
#' A data frame with the same number of rows as `x`. The first column (`term`)
#' contains the original values. Subsequent columns contain the match result
#' for each match type (e.g. `exact_match`, `case_insensitive_match`,
#' `fuzzy_boundary_match`, `fuzzy_encoding_match`). Rows for NA values in `x`
#' are all NAs.
#'
#' @export
#'
#' @examples
#' data(colour_thesaurus)
#'
#' x <- c("red", "lipstick", "green", "mint", "blue", "azure")
#' control_matches(x, colour_thesaurus)
control_matches <- function(x, thesaurus,
                            thesaurus_cols = c(1, 2),
                            case_insensitive = FALSE,
                            fuzzy_boundary = FALSE,
                            fuzzy_encoding = FALSE) {
  if (!is.vector(x)) {
    rlang::abort("`x` must be a vector.")
  }

  terms <- unique(x[!is.na(x)])
  thesaurus <- prepare_thesaurus(thesaurus, thesaurus_cols)

  matched <- match_terms(terms, thesaurus, case_insensitive,
                         fuzzy_boundary, fuzzy_encoding)

  # Rename match columns to add _match suffix
  match_cols <- setdiff(names(matched), "term")
  names(matched)[names(matched) %in% match_cols] <- paste0(match_cols, "_match")

  result <- matched[match(x, matched$term), , drop = FALSE]
  row.names(result) <- NULL
  result
}

test_that("control_matches() preserves NAs in x", {
  x <- c("red", NA, "blue")
  result <- control_matches(x, colour_thesaurus)

  expect_equal(nrow(result), 3)
  expect_true(all(is.na(result[2, ])))
  expect_true(all(!is.na(result[1, ])))
  expect_true(all(!is.na(result[3, ])))
})

test_that("control_matches() returns a data frame with correct columns", {
  x <- c("red", "lipstick", "green")
  result <- control_matches(x, colour_thesaurus)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 3)
  expect_equal(ncol(result), 2)
  expect_true("term" %in% names(result))
  expect_true("exact_match" %in% names(result))
})

test_that("control_matches() includes term column with original values", {
  x <- c("red", "lipstick", "green")
  result <- control_matches(x, colour_thesaurus)

  expect_equal(result$term, x)
})

test_that("control_matches() returns columns for each active match type", {
  x <- c("RED", "lipstick")
  result <- control_matches(x, colour_thesaurus, case_insensitive = TRUE)

  expect_equal(ncol(result), 3)
  expect_true("term" %in% names(result))
  expect_true("exact_match" %in% names(result))
  expect_true("case_insensitive_match" %in% names(result))
})

test_that("control_matches() works with exact matching", {
  result <- control_matches(colour_thesaurus$shade, colour_thesaurus)

  expect_equal(result$exact_match, colour_thesaurus$colour)
})

test_that("control_matches() works with case insensitive matching", {
  result <- control_matches(toupper(colour_thesaurus$shade), colour_thesaurus,
                            case_insensitive = TRUE)

  expect_equal(result$case_insensitive_match, colour_thesaurus$colour)
})

test_that("control_matches() works with fuzzy matching", {
  df <- data.frame(canon = "foo bar", variant = "foo-bar")
  result <- control_matches(c("foo bar", "foo_bar"), df, fuzzy_boundary = TRUE)

  expect_equal(result$fuzzy_boundary_match, c("foo bar", "foo bar"))
})

test_that("control_matches() aborts if `x` is not a vector", {
  expect_error(control_matches(data.frame(), colour_thesaurus),
               regexp = "`x` must be a vector")
})

# data.R
# Dataset documentation

#' Colour thesaurus
#'
#' A dataset of colour names based on Ingrid Sundberg's "colour thesaurus".
#' Contains 240 names for shades of 12 colours, formatted for use as a thesaurus
#' in [control()].
#'
#' @format A data frame with 240 rows and 2 variables:
#' \describe{
#'   \item{colour}{preferred colour name: `"white"`, `"tan"`, `"yellow"`,
#'   `"orange"`, `"red"`, `"pink"`, `"purple"`, `"blue"`, `"green"`, `"brown"`,
#'   `"grey"`, or `"black"`}
#'   \item{shade}{variant names for shades of `colour`}
#' }
#' @source 
#' Ingrid Sundberg (\url{https://web.archive.org/web/20250619164626/https://ingridsundberg.com/2014/02/04/the-color-thesaurus/})
"colour_thesaurus"

# control_names.R
# Control names of an object

#' Control the names of an object
#'
#' @description
#' `control_names()` controls the names of an object using a thesaurus.
#' It extracts the names of `x`, passes them to [control()], and reassigns
#' the result to `names(x)`.
#'
#' `control_names_ci()` and `control_names_fuzzy()` are convenience aliases
#' for case insensitive and full fuzzy matching respectively.
#'
#' @param x Object with names to control.
#' @param thesaurus Data frame with a vector of preferred terms and a vector of
#'   variants.
#' @param thesaurus_cols Vector of two column names or positions specifying which
#'   columns in `thesaurus` contain the preferred terms and variants
#'   respectively. Defaults to the first two columns.
#' @param ... Other arguments passed to [control()]. This includes fuzzy
#'   matching options (`case_insensitive`, `fuzzy_boundary`, `fuzzy_encoding`)
#'   and output options (`quiet`, `warn_unmatched`).
#'
#' @return
#' The object `x` with its names replaced by controlled values.
#'
#' @export
#'
#' @examples
#' df <- data.frame(temp = 20, humid = 65, `wind speed` = 10, date = "2024-01-01")
#' thesaurus <- data.frame(
#'   preferred = c("temperature", "humidity", "wind_speed"),
#'   variant = c("temp", "humid", "wind speed")
#' )
#' control_names(df, thesaurus)
control_names <- function(x, thesaurus, thesaurus_cols = c(1, 2), ...) {
  if (is.null(names(x))) {
    rlang::abort("`x` must have names.")
  }
  names(x) <- control(names(x), thesaurus, thesaurus_cols, ...)
  x
}

#' @rdname control_names
#' @export
control_names_ci <- function(x, thesaurus, thesaurus_cols = c(1, 2), ...) {
  control_names(x, thesaurus, thesaurus_cols, case_insensitive = TRUE, ...)
}

#' @rdname control_names
#' @export
control_names_fuzzy <- function(x, thesaurus, thesaurus_cols = c(1, 2), ...) {
  control_names(x, thesaurus, thesaurus_cols, case_insensitive = TRUE,
                fuzzy_boundary = TRUE, fuzzy_encoding = TRUE, ...)
}

test_that("control_names() aborts if `x` has no names", {
  expect_error(control_names(1:3, colour_thesaurus),
               regexp = "`x` must have names")
})

test_that("control_names() with exact matching works", {
  df <- data.frame(lipstick = 1, mint = 2, azure = 3)
  result <- control_names(df, colour_thesaurus, quiet = TRUE, warn_unmatched = FALSE)

  expect_equal(names(result), c("red", "green", "blue"))
  expect_equal(result$red, 1)
  expect_equal(result$green, 2)
  expect_equal(result$blue, 3)
})

test_that("control_names_ci() with case insensitive matching works", {
  df <- data.frame(LIPSTICK = 1, MINT = 2, AZURE = 3)
  result <- control_names_ci(df, colour_thesaurus, quiet = TRUE, warn_unmatched = FALSE)

  expect_equal(names(result), c("red", "green", "blue"))
})

test_that("control_names_fuzzy() with fuzzy matching works", {
  df <- data.frame(foo_bar = 1, foobar = 2)
  thesaurus <- data.frame(preferred = "foo bar", variant = "foo-bar")
  result <- control_names_fuzzy(df, thesaurus, quiet = TRUE, warn_unmatched = FALSE)

  expect_equal(names(result), c("foo bar", "foo bar"))
})

test_that("control_names() preserves object structure", {
  df <- data.frame(a = 1, b = 2, c = 3)
  thesaurus <- data.frame(preferred = "x", variant = "a")
  result <- control_names(df, thesaurus, quiet = TRUE, warn_unmatched = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(ncol(result), 3)
  expect_equal(nrow(result), 1)
  expect_equal(result[[1]], 1)
})

# control_helpers.R
# Internal helpers for control verbs

validate_thesaurus <- function(thesaurus) {
  if (ncol(thesaurus) != 2) {
    rlang::abort("`thesaurus_cols` must specify exactly 2 columns.")
  }

  if (anyDuplicated(thesaurus[[2]]) > 0) {
    rlang::abort("Variants (column 2 of `thesaurus`) must be unique.")
  }

  invisible(thesaurus)
}

prepare_thesaurus <- function(thesaurus, thesaurus_cols) {
  thesaurus <- as.data.frame(thesaurus)[, thesaurus_cols, drop = FALSE]
  validate_thesaurus(thesaurus)
  names(thesaurus) <- c("canon", "variant")
  thesaurus
}

fuzzy_match <- function(terms, variants, canon, char_class, ignore_case = FALSE) {
  patterns <- paste0("^", gsub(char_class, ".?", variants, perl = TRUE), "$")
  vapply(terms, function(term) {
    idx <- which(vapply(patterns, grepl, logical(1), x = term, ignore.case = ignore_case, perl = TRUE))
    if (length(idx) > 0) canon[idx[1]] else NA_character_
  }, character(1))
}

match_terms <- function(terms, thesaurus,
                        case_insensitive = FALSE,
                        fuzzy_boundary = FALSE,
                        fuzzy_encoding = FALSE) {
  result <- data.frame(term = terms)

  result$exact <- thesaurus$canon[match(result$term, thesaurus$variant)]

  if (isTRUE(case_insensitive)) {
    result$case_insensitive <- thesaurus$canon[match(tolower(result$term),
                                                     tolower(thesaurus$variant))]
  }

  if (isTRUE(fuzzy_boundary)) {
    result$fuzzy_boundary <- fuzzy_match(result$term,
                                         thesaurus$variant,
                                         thesaurus$canon,
                                         "[[:punct:][:space:]]",
                                         ignore_case = case_insensitive)
  }

  if (isTRUE(fuzzy_encoding)) {
    result$fuzzy_encoding <- fuzzy_match(result$term,
                                         thesaurus$variant,
                                         thesaurus$canon,
                                         "[^[:ascii:]]",
                                         ignore_case = case_insensitive)
  }

  result
}

# fish.R
# Functions for Historic England's FISH controlled vocabularies
# <https://heritage-standards.org.uk/fish-vocabularies/>

#' Read a FISH controlled vocabulary
#'
#' Reads controlled vocabularies from Historic England's FISH (Forum on
#' Information Standards in Heritage) group. These can be downloaded from:
#' <https://heritage-standards.org.uk/fish-vocabularies/>.
#'
#' @param path Path or URL to a vocabulary in FISH's CSV format. Can be
#'   either a `.zip` archive or a directory containing already uncompressed
#'   files, or a URL.
#'
#' @return
#' A data frame with two columns: `preferred` and `term`.
#'
#' @references
#' * Forum on Information Standards in Heritage, "FISH Thesaurus Table Structure",
#'   Available from: <https://heritage-standards.org.uk/fish-vocabularies/>.
#'
#' @export
#'
#' @examples
#' # Read a FISH vocabulary from a local zip file
#' nationality_zip <- system.file("extdata", "fish-nationality.zip",
#'                                package = "controller")
#' read_fish(nationality_zip)
read_fish <- function(path) {
  cleanup <- character()
  on.exit(unlink(cleanup, recursive = TRUE), add = TRUE)

  # Download if URL (don't check extension)
  if (is_url(path)) {
    tmp <- fs::file_temp(ext = "zip")
    cleanup <- c(cleanup, tmp)
    path <- curl::curl_download(path, tmp)
  }

  # Determine file list
  if (fs::file_exists(path) && is_zip(path)) {
    tmp_dir <- fs::file_temp()
    fs::dir_create(tmp_dir)
    cleanup <- c(cleanup, tmp_dir)
    files <- utils::unzip(path, exdir = tmp_dir)
  }
  else if (fs::file_exists(path) && fs::is_dir(path)) {
    files <- fs::dir_ls(path)
  }
  else {
    rlang::abort(
      "`path` must be a path to an existing .zip file or directory, or a URL.",
      class = "controller_read_error"
    )
  }

  # Normalize file names
  file_names <- fs::path_ext_remove(fs::path_file(files))
  normalized <- gsub("_", "", tolower(file_names))
  names(files) <- normalized

  # Validate required files
  required <- c("thesaurusterms", "thesaurustermpreferences")
  missing <- setdiff(required, names(files))
  if (length(missing) > 0) {
    rlang::abort(
      c("FISH vocabulary is missing required files.",
        i = paste("Expected:", paste(required, collapse = ", ")),
        x = paste("Found:", paste(names(files), collapse = ", "))),
      class = "controller_read_error"
    )
  }

  # Read data
  terms <- utils::read.csv(files[["thesaurusterms"]])
  preferred <- utils::read.csv(files[["thesaurustermpreferences"]])

  data.frame(
    preferred = terms$TERM[match(preferred$THE_TE_UID_2, terms$THE_TE_UID)],
    term = terms$TERM[match(preferred$THE_TE_UID_1, terms$THE_TE_UID)]
  )
}

test_that("read_fish() reads from zip file", {
  zip_path <- system.file("extdata", "fish-nationality.zip",
                          package = "controller")
  result <- read_fish(zip_path)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("preferred", "term"))
  expect_gt(nrow(result), 0)
})

test_that("read_fish() reads from directory", {
  zip_path <- system.file("extdata", "fish-nationality.zip",
                          package = "controller")
  temp_dir <- fs::file_temp()
  fs::dir_create(temp_dir)
  on.exit(fs::dir_delete(temp_dir), add = TRUE)
  utils::unzip(zip_path, exdir = temp_dir)

  result <- read_fish(temp_dir)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("preferred", "term"))
  expect_gt(nrow(result), 0)
})

test_that("read_fish() errors on invalid path", {
  expect_error(
    read_fish("/nonexistent/path/to/file.zip"),
    class = "controller_read_error"
  )
})

test_that("read_fish() errors on non-zip file", {
  temp_file <- fs::file_temp(ext = "zip")
  writeLines("not a zip file", temp_file)
  on.exit(fs::file_delete(temp_file), add = TRUE)

  expect_error(
    read_fish(temp_file),
    class = "controller_read_error"
  )
})

test_that("read_fish() errors on zip missing required files", {
  temp_dir <- fs::file_temp()
  fs::dir_create(temp_dir)
  on.exit(fs::dir_delete(temp_dir), add = TRUE)

  writeLines("a,b\n1,2", fs::path(temp_dir, "wrong_file.csv"))
  temp_zip <- fs::file_temp(ext = "zip")
  utils::zip(temp_zip, files = fs::path(temp_dir, "wrong_file.csv"), flags = "-r9Xq")
  on.exit(fs::file_delete(temp_zip), add = TRUE)

  expect_error(
    read_fish(temp_zip),
    class = "controller_read_error"
  )
})

---
title: "Controlled vocabularies"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Controlled vocabularies}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

A *controlled vocabulary* is a predefined set of terms intended to be used to be used in a specific context.
They can solve a number of problems of data standardisation and cleaning in data science.
controller provides functions for working with controlled vocabularies in R. 

This vignette explains the main features of the package.
It introduces the `control()` verb, which recodes values in a vector using a lookup table of preferred and variant terms (a *thesaurus*).

```{r setup, include = FALSE}
library(controller)
```

## The `control()` verb

A common data-tidying problem is standardising variant terms for the same concept.
Imagine we have a dataset that uses a number of different names for shades of the same colour.
As data analysts, we naturally want to recode the data to eliminate this messy creativity, for example using [dplyr::recode()](https://dplyr.tidyverse.org/reference/recode.html):

```{r eg-dplyr}
library(dplyr, warn.conflicts = FALSE)
shades <- c("daffodil", "purple", "magenta", "azure", "navy", "violet")

recode(shades,
       daffodil = "yellow",
       purple = "purple",
       magenta = "pink",
       azure = "blue",
       navy = "blue",
       violet = "purple")
```

But recoding this way can be tedious, especially if there are a large number of terms.
With `control()`, we can instead use a data frame containing a thesaurus to replace the values:

```{r eg-control}
library(controller)
data("colour_thesaurus")

control(shades, colour_thesaurus)
```

By default, `control()` issues a message listing the values that were replaced and a warning for any values that could not be matched in the thesaurus.
These can be suppressed with `quiet = TRUE` and `warn_unmatched = FALSE` respectively.

Use `control_names()` to control the names of an object rather than its values.
This is useful for standardising column names in data frames:

```{r eg-control-names}
df <- data.frame(temp = 20, humid = 65, `wind speed` = 10, date = "2024-01-01")
df

control_names(df, thesaurus = data.frame(
  preferred = c("temperature", "humidity", "wind_speed"),
  variant = c("temp", "humid", "wind speed")
))
```

## Fuzzy matching

`control()` also supports fuzzy matching, removing the need to exhaustively list variants for common causes of differing terminology.
For example, to perform a case insensitive match to the thesaurus:

```{r eg-ci}
shades_ci <- toupper(shades)
control_ci(shades_ci, colour_thesaurus)
```

`control_fuzzy()` goes further by also ignoring differences in word boundaries and character encoding.
Fuzzy boundary matching treats word boundaries like spaces, hyphens, and underscores as equivalent:

```{r eg-boundary}
df <- data.frame(preferred = "foo bar", variant = "foo-bar")
control(c("foo bar", "foo_bar", "foobar"), df,
        fuzzy_boundary = TRUE, quiet = FALSE)
```

Fuzzy encoding matching ignores non-ASCII characters, so plain ASCII input can match variants with diacritics or wrongly encoded characters ('mojibake'):

```{r eg-encoding}
df <- data.frame(preferred = "bar", variant = "fo\u00f6")
control("foo", df, fuzzy_encoding = TRUE, quiet = FALSE)
```

To inspect which type of match was used for each value, use `control_matches()`.
It returns a data frame with a column for each match type, rather than a single vector:

```{r eg-coalesce}
control_matches(shades_ci, colour_thesaurus, case_insensitive = TRUE)
```

## Thesaurus format

A thesaurus is a data frame with two columns containing preferred terms and the variants to be controlled.
Each row maps one variant to one preferred term.
The variants must be unique values – otherwise they cannot be unambiguously matched to a preferred term.

`colour_thesaurus`,^[Compiled by Ingrid Sundberg, <https://web.archive.org/web/20250619164626/https://ingridsundberg.com/2014/02/04/the-color-thesaurus/>.] used in the examples above, is an example of a thesaurus bundled with the package:

```{r thesaurus-preview}
head(colour_thesaurus)
```

Because a thesaurus is simply a data frame, it can be created or read into R with standard tools:

```{r eg-custom-thesaurus}
my_thesaurus <- data.frame(
  preferred = c("feline", "canine"),
  variant = c("cat", "dog")
)

animals <- c("cat", "dog", "parrot")
control(animals, my_thesaurus)
```

Use the `thesaurus_cols` argument to specify which columns of the data frame contain the preferred and variant terms respectively:

```{r eg-thesaurus-cols}
my_wider_thesaurus <- data.frame(
  code = c("FEL", "CAN"),
  label = c("feline", "canine"),
  variant = c("cat", "dog")
)

control(animals, my_wider_thesaurus, thesaurus_cols = c("label", "variant"))
```

The package can also read controlled vocabularies in Historic England's [FISH](https://heritage-standards.org.uk/fish-vocabularies/) (Forum on Information Standards in Heritage) format using `read_fish()`.
The `path` argument can be a local `.zip` file, an uncompressed directory, or a URL:

```{r eg-fish}
nationality_zip <- system.file("extdata", "fish-nationality.zip",
                               package = "controller")
fish_nationality <- read_fish(nationality_zip)
head(fish_nationality)
```

The result is a data frame with columns `preferred` and `term`, ready to use with `control()`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{control}
\alias{control}
\alias{control_ci}
\alias{control_fuzzy}
\title{Recode values using a thesaurus}
\usage{
control(
  x,
  thesaurus,
  thesaurus_cols = c(1, 2),
  case_insensitive = FALSE,
  fuzzy_boundary = FALSE,
  fuzzy_encoding = FALSE,
  quiet = FALSE,
  warn_unmatched = TRUE
)

control_ci(x, thesaurus, thesaurus_cols = c(1, 2), ...)

control_fuzzy(x, thesaurus, thesaurus_cols = c(1, 2), ...)
}
\arguments{
\item{x}{Vector to recode.}

\item{thesaurus}{Data frame with a vector of preferred terms and a vector of
variants.}

\item{thesaurus_cols}{Vector of two column names or positions specifying which
columns in \code{thesaurus} contain the preferred terms and variants
respectively. Defaults to the first two columns.}

\item{case_insensitive}{Set to \code{TRUE} to perform case insensitive matching.}

\item{fuzzy_boundary}{Set to \code{TRUE} to perform fuzzy matching that ignores
differences in the word boundaries used (e.g. \code{"foo bar"} matches \code{"foo-bar"}).}

\item{fuzzy_encoding}{Set to \code{TRUE} to perform fuzzy matching that ignores
non-ASCII characters that may have been encoded differently (e.g. \code{"foo"}
matches \code{"foö"}).}

\item{quiet}{Set to \code{TRUE} suppress messages about replaced values.}

\item{warn_unmatched}{If \code{TRUE} (the default), issues a warning for values
that couldn't be matched in \code{thesaurus}.}

\item{...}{For \code{control_ci()} and \code{control_fuzzy()}, other arguments passed
to \code{control()}. This includes fuzzy matching options (\code{fuzzy_boundary},
\code{fuzzy_encoding}) and output options (\code{quiet}, \code{warn_unmatched}).}
}
\value{
A vector the same length as \code{x} with values matching variants in \code{thesaurus}
replaced with the preferred term. NAs in \code{x} are preserved as NAs.

By default gives a message listing replaced values and a warning listing any
values not matched in the thesaurus. These can be suppressed with
\code{quiet = TRUE} and \code{warn_unmatched = FALSE} respectively.
}
\description{
The \code{control()} verb replaces values in a vector with values looked up in a
thesaurus. It is similar to \code{\link[=switch]{switch()}} or \code{\link[dplyr:recode]{dplyr::recode()}}, but the
replacement values are specified as a data frame instead of as individual
arguments.

By default \code{control()} replaces only values of \code{x} that exactly match terms
in \code{thesaurus}. Additional arguments allow for case insensitive and fuzzy
matching strategies (see details). \code{control_ci()} and \code{control_fuzzy()} are
convenience aliases for case insensitive exact matching and full fuzzy
matching respectively.
}
\examples{
data(colour_thesaurus)

# Exact matching
x <- c("red", "lipstick", "green", "mint", "blue", "azure")
control(x, colour_thesaurus)

# Case insensitive matching
x <- toupper(x)
control_ci(x, colour_thesaurus)

# control_matches() returns a data frame showing which match type was used:
control_matches(x, colour_thesaurus, case_insensitive = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control_matches.R
\name{control_matches}
\alias{control_matches}
\title{Show detailed match results from a thesaurus}
\usage{
control_matches(
  x,
  thesaurus,
  thesaurus_cols = c(1, 2),
  case_insensitive = FALSE,
  fuzzy_boundary = FALSE,
  fuzzy_encoding = FALSE
)
}
\arguments{
\item{x}{Vector to recode.}

\item{thesaurus}{Data frame with a vector of preferred terms and a vector of
variants.}

\item{thesaurus_cols}{Vector of two column names or positions specifying which
columns in \code{thesaurus} contain the preferred terms and variants
respectively. Defaults to the first two columns.}

\item{case_insensitive}{Set to \code{TRUE} to perform case insensitive matching.}

\item{fuzzy_boundary}{Set to \code{TRUE} to perform fuzzy matching that ignores
differences in the word boundaries used (e.g. \code{"foo bar"} matches \code{"foo-bar"}).}

\item{fuzzy_encoding}{Set to \code{TRUE} to perform fuzzy matching that ignores
non-ASCII characters that may have been encoded differently (e.g. \code{"foo"}
matches \code{"foö"}).}
}
\value{
A data frame with the same number of rows as \code{x}. The first column (\code{term})
contains the original values. Subsequent columns contain the match result
for each match type (e.g. \code{exact_match}, \code{case_insensitive_match},
\code{fuzzy_boundary_match}, \code{fuzzy_encoding_match}). Rows for NA values in \code{x}
are all NAs.
}
\description{
\code{control_matches()} returns a data frame showing which type of match was used
for each value in \code{x}. This is useful for debugging or inspecting how
\code{control()} recodes values.
}
\examples{
data(colour_thesaurus)

x <- c("red", "lipstick", "green", "mint", "blue", "azure")
control_matches(x, colour_thesaurus)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control_names.R
\name{control_names}
\alias{control_names}
\alias{control_names_ci}
\alias{control_names_fuzzy}
\title{Control the names of an object}
\usage{
control_names(x, thesaurus, thesaurus_cols = c(1, 2), ...)

control_names_ci(x, thesaurus, thesaurus_cols = c(1, 2), ...)

control_names_fuzzy(x, thesaurus, thesaurus_cols = c(1, 2), ...)
}
\arguments{
\item{x}{Object with names to control.}

\item{thesaurus}{Data frame with a vector of preferred terms and a vector of
variants.}

\item{thesaurus_cols}{Vector of two column names or positions specifying which
columns in \code{thesaurus} contain the preferred terms and variants
respectively. Defaults to the first two columns.}

\item{...}{Other arguments passed to \code{\link[=control]{control()}}. This includes fuzzy
matching options (\code{case_insensitive}, \code{fuzzy_boundary}, \code{fuzzy_encoding})
and output options (\code{quiet}, \code{warn_unmatched}).}
}
\value{
The object \code{x} with its names replaced by controlled values.
}
\description{
\code{control_names()} controls the names of an object using a thesaurus.
It extracts the names of \code{x}, passes them to \code{\link[=control]{control()}}, and reassigns
the result to \code{names(x)}.

\code{control_names_ci()} and \code{control_names_fuzzy()} are convenience aliases
for case insensitive and full fuzzy matching respectively.
}
\examples{
df <- data.frame(temp = 20, humid = 65, `wind speed` = 10, date = "2024-01-01")
thesaurus <- data.frame(
  preferred = c("temperature", "humidity", "wind_speed"),
  variant = c("temp", "humid", "wind speed")
)
control_names(df, thesaurus)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/controller-package.R
\docType{package}
\name{controller-package}
\alias{controller}
\alias{controller-package}
\title{controller: Controlled Vocabularies}
\description{
Work with controlled vocabularies; a predefined set of terms intended to be used in in a specific context. This package introduces the 'control' verb, which recodes values in a vector using a lookup table of preferred and variant terms (a 'thesaurus'). It also includes functions for reading controlled vocabularies from standard formats.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://controller.joeroe.io}
  \item \url{https://github.com/joeroe/controller}
  \item Report bugs at \url{https://github.com/joeroe/controller/issues}
}

}
\author{
\strong{Maintainer}: Joe Roe \email{joe@joeroe.io} (\href{https://orcid.org/0000-0002-1011-1244}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Joe Roe \email{joe@joeroe.io} (\href{https://orcid.org/0000-0002-1011-1244}{ORCID}) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{colour_thesaurus}
\alias{colour_thesaurus}
\title{Colour thesaurus}
\format{
A data frame with 240 rows and 2 variables:
\describe{
\item{colour}{preferred colour name: \code{"white"}, \code{"tan"}, \code{"yellow"},
\code{"orange"}, \code{"red"}, \code{"pink"}, \code{"purple"}, \code{"blue"}, \code{"green"}, \code{"brown"},
\code{"grey"}, or \code{"black"}}
\item{shade}{variant names for shades of \code{colour}}
}
}
\source{
Ingrid Sundberg (\url{https://web.archive.org/web/20250619164626/https://ingridsundberg.com/2014/02/04/the-color-thesaurus/})
}
\usage{
colour_thesaurus
}
\description{
A dataset of colour names based on Ingrid Sundberg's "colour thesaurus".
Contains 240 names for shades of 12 colours, formatted for use as a thesaurus
in \code{\link[=control]{control()}}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fish.R
\name{read_fish}
\alias{read_fish}
\title{Read a FISH controlled vocabulary}
\usage{
read_fish(path)
}
\arguments{
\item{path}{Path or URL to a vocabulary in FISH's CSV format. Can be
either a \code{.zip} archive or a directory containing already uncompressed
files, or a URL.}
}
\value{
A data frame with two columns: \code{preferred} and \code{term}.
}
\description{
Reads controlled vocabularies from Historic England's FISH (Forum on
Information Standards in Heritage) group. These can be downloaded from:
\url{https://heritage-standards.org.uk/fish-vocabularies/}.
}
\examples{
# Read a FISH vocabulary from a local zip file
nationality_zip <- system.file("extdata", "fish-nationality.zip",
                               package = "controller")
read_fish(nationality_zip)
}
\references{
\itemize{
\item Forum on Information Standards in Heritage, "FISH Thesaurus Table Structure",
Available from: \url{https://heritage-standards.org.uk/fish-vocabularies/}.
}
}
